#!/usr/local/bin/perl
##########################################################################
#
# Revision 4 of Mail Filter Relay, written by RPD Mallett, Vine Micros Ltd
# 23rd August 2002.  Freeware.
#
# Original sock-handling code from examples on web:
# by Prof. Golden G. Richard III, Dept. of 
# Computer Science, University of New Orleans, April 1996-March 1998.
#
# E-Mail sending bits from Novell samples.
#
# This is beta-test code - I make no warranties as to its suitability for
# any purpose as it was written by someone with only 2 days' experience
# in Perl.  If you're not confident in running it, don't!  It's a very
# simple Perl program, but adds a really neat function to BM.  Don't run
# it using any other Mail Proxy, or you may get unexpected results.
# I've been running it for months now, with no problems* - but that's
# no guarantee that it won't have them!
#
# Its function is to act as a relay between BM's incoming-mail Mail Proxy
# and your internal mail server, using SMTP.  It checks all incoming mail
# against a filter list (sys:\filtlist.txt).  Enter annoying words or phrases
# in filtlist.txt, each on a separate line (don't add blank lines!).  Only
# use alphanumeric characters, as some (eg. '.') may be intepreted as Perl
# expressions.  Put spaces either side of short words to prevent it from
# filtering emails with longer words that contain the shorter word.
# (eg. 'Middlesex' - a county in England - contains the word 'sex', so use
# ' sex ' instead, without the quotes of course).  
#
# A sample filter list could be (ignore the #'s at the beginning, noting
# any spaces I've added, and not leaving any lines blank!):
#
# sex 
#\@clinton\.com
#name=.+\.exe
#nakedwife
#
# The first example should be clear.  The second is odd because Perl
# uses the @ and . characters for other purposes, so you have to add the
# backslash before them.  The same for .exe, which should filter out any
# Executable files being sent to you.  Once you get to know Perl, you could
# do some fancy things with the filter list and various special characters.
#
# Any email that matches is dumped to sys:\filtered.log, with a summary of
# why in sys:\filtword.log.  It's very easy to accidentally filter a valid
# email, so keep an eye on the filter logs.
#
# Instructions:
# 1. Save this code to sys:\perl\scripts
# 2. Make sure Perl 5 is present on your server (eg. as on Netware 5.0).
# 3. Change the 'orac4' bit in this code to your email server's domain name.
# 4. Add a secondary IP address, if there's already something on port 25
#    (such as BorderManager's Mail Proxy!).  To do this type:
#	add secondary ipaddress 10.1.2.4
#    (Change the 10.1.2.4 to whatever IP address is suitable for your 
#    server.  Eg. If your server is 192.168.0.1 then perhaps use
#    192.168.0.2)
# 5. Change the IP address in this code from "\x0A\x01\x02\x04" to the new
#    secondary IP address - in hexadecimal (mine is 10.1.2.4 in decimal).
# 6. Change the 'relay\@vinemicros.com' to suit your own domain name.
# 7. Point your Bordermanager Mail Relay to this IP address instead of to
#    your internal mail server.
# 8. Run this code (perl relay.pl).
#
# I recommend adding the following lines to your AUTOEXEC.NCF file:
#	add secondary ipaddress 10.1.2.4
#	perl relay.pl
# This will make the code run each time the server is re-started.  You
# should make sure that this runs BEFORE Bordermanager is started, as the
# first thing BM does is send an alert email (via this relay) to tell you
# that it's been loaded.  If the relay isn't loaded first, then BM may have
# problems.
#
#
# Revision history:
#
# Version 4: Removed bug that caused mail filter to halt when receiving
# attachments with no text after them - ie. those where the MIME-encoded
# file abruptly ended.
# This version also now shows the time & date of the last email processed.
#
# Version 3 has a speed improvement.  Emails with big attachements took
# a long time to scan - even though the MIME-encoded attachment is full
# of random-looking text and therefore shouldn't need scanning.  This
# version has basic MIME-detection included (ie. any email line with 70 
# characters or more, none of which are spaces).  If 10 or more of these
# are detected in succession, then the following MIME lines are skipped.
# This avoids long delays in scanning emails with attachements.
# I've also put together all the bits that you need to change in one area.
#
# Version 2 now sends MAIL FROM <relay@vinemicros.com>, as it was found
# that some (valid) emails were missing the MAIL FROM email address, and
# this confused the internal email server.
#
# Any hints or suggestions to:  richard@vinemicros.com


use Socket;

$port = 25;
($name, $aliases, $protocol) = getprotobyname('tcp');
if ($port !~ /^\d+$/) {
     ($name, $aliases, $port) = getservbyport($port, 'tcp');
}

print "Listening on port $port...\n";

socket(S,AF_INET,SOCK_STREAM,$protocol) || die "socket : $!";
$sockaddr = 'S n a4 x8';

# ------------------------ Here are the bits to change ------------------------------
# Change "\x0A\x01\x02\x04" to suit the secondary IP address you added.
# eg. if it's 192.168.0.1 it would be "\xC0\xA8\x00\x01"
# This is the IP address that this program 'listens' on.
$this = pack($sockaddr, AF_INET, $port, "\x0A\x01\x02\x04");	# IP address, in hex.

# Change this to your email server's domain name
# It might be worth doing a ping (eg. ping orac4) from the server console to make
# sure you've entered a valid domain name, and that it resolves correctly.
my($mailServer) = 'orac4';

# Change this to suit your requirements.  It's just the email address that your
# email server will be told the email is from.
$relay='relay@vinemicros.com';
# ------------------------ End of bits to change ------------------------------------


bind(S, $this) || die "bind : $!";
# created 10 queues to listen on - although only one is ever active at a time.
listen(S,10) || die "listen: $!";
select(S);
$| = 1;
select(STDOUT);

# Create connections as clients "arrive".  $con maintains the connection 
# number of the last client

$dumped=0;

for ($con = 1; ; $con++) {
        printf "\n".localtime();
	printf "\nWaiting for connection $con (filtered $dumped)\n";

	($addr = accept(NS,S)) || die $!; 
	select(NS);
	$| = 1;

	select(STDOUT);

	($af,$port, $inetaddr) = unpack($sockaddr, $addr);
	@inetaddr = unpack('C4', $inetaddr);

	print "Serving connection @ Internet address @inetaddr\n";

	$helo="";
	$mailfrom="";
	@rcptto=("");
	$rcpttoqty=0;
	$data="";
	@content=("");
	$contentqty=0;
	$quit="";
	@badword=("");
	$badwordqty=0;

	$dataactive=0;
	$last="";
	$crlf="\x0D\x0A";
	$cr="\x0D";
	$mime=0;


	open(FILTER, "sys:/filtlist.txt") or die "Can't open filtlist.txt";
	@filtlist=(<FILTER>,"To: fax#");
	chomp(@filtlist);
	close(FILTER);

	print NS "220 Relay filter ready$crlf";

	while (<NS>) {
		$in=$_;
		$reply="";

		if ($dataactive==0) {
			print "Incoming: $_";
			$last="";

			if (m/^HELO/i) {
				$helo=$_;
				$reply="250 OK$crlf";
			}

			if (m/^MAIL FROM:/i) {
				$mailfrom=$_;
				$reply="250 OK$crlf";
			}

			if (m/^RCPT TO:/i) {
				@rcptto[$rcpttoqty]=$_;
				$rcpttoqty++;
				$reply="250 OK$crlf";
			}

			if (m/^DATA/i) {
				$data=$_;
				$reply="354 Go ahead$crlf";
				$dataactive=1;
			}

			if (m/^QUIT/i) {
				$quit=$_;
				$reply="221 Closing connection$crlf";
			}
		}
		else {
			@content[$contentqty]=$_;
			$contentqty++;

			$_=$last.$in;
			# MIME detection
			# We add together the last two lines, in case a word-wrap has split a MIME-line into two.
			if (length($_)>70)	{
				if (m/ /i) {
					$mime=0;
				}
				else {
					$mime=$mime+1;
				}
			}
			else {
				$mime=0;
			}			

			if ($mime<10) {
				foreach $filt (@filtlist) {
					if ($in =~ m/$filt/i) {
						print "Badword detected: $filt\n";
						@badword[$badwordqty]=$filt;
						$badwordqty++;
					}
				}
				print "Line $contentqty scanned         $cr";
	
			}
			else {
				print "Line $contentqty MIME-skipped    $cr";
			}

			if (m/\015\012\x2E\015\012/m) {
				print "\nFound end of text.\n";
				$dataactive=0;
				$reply="250 OK$crlf";
			}
		}

		if ($reply) {
			print NS $reply;
		}


		$last=$in;
	}


	close(NS);


	if ($badwordqty == 0) {
		# These constants shouldn't need changing.
		my($packFormat) = 'S n a4 x8';
		my($proto)      = getprotobyname("tcp")        || 6;
		my($port)       = getservbyname("SMTP", "tcp") || 25;
		my($name,$aliases,$addrtype,$len,@addrs) = gethostbyname($mailServer);
		my($serverAddr) = $addrs[0];

		if (! defined($len)) {
			die('gethostbyname failed.');
		}

		socket(SMTP, AF_INET(), SOCK_STREAM(), $proto)
			or die("socket: $!");

		connect(SMTP, pack($packFormat, AF_INET(), $port, $serverAddr))
			or die("connect: $!");

		select(SMTP); $| = 1; select(STDOUT);    # use unbuffered i/o.

		recv(SMTP, $_, 200, 0);		# Read Mail server's start up message
		sendSMTP(1, $helo);
		sendSMTP(1, "MAIL FROM: <$relay>\n");
		foreach (@rcptto) {
			sendSMTP(1, $_);
		}
		sendSMTP(1, $data);
		foreach (@content) {
			send(SMTP, $_, 0);
		}
		sendSMTP(0, "");		# Send blank line, to get a response.
		sendSMTP(1, $quit);

		close(SMTP);
	}

	else {
		$dumped++;

		open(LOG, ">>sys:/filtword.log");
		print LOG "\n".localtime();
		foreach (@badword) {
			print LOG "\n$_";
		}
		print LOG "\n$mailfrom";
		foreach (@rcptto) {
			print LOG "$_";
		}
		close(LOG);		

		open(LOG, ">>sys:/filtered.log");
		print LOG "\n\n\n$helo$mailfrom";
		foreach (@rcptto) {
			print LOG "$_";
		}
		print LOG "$data";
		foreach (@content) {
			print LOG "$_";
		}
		close(LOG);		
	}

} 



  sub closeSocket {     # close smtp socket on error
     close(SMTP);
     die("SMTP socket closed due to SIGINT\n");
  }

  sub send_to_smtp {
     my($debug)  = shift;
     my($response) = shift;
     my($buffer) = @_;

     print STDERR ("> $buffer") if $debug;
     send(SMTP, $buffer, 0);

     if ($response) {
       recv(SMTP, $buffer, 200, 0);
       print STDERR ("< $buffer");
       return( (split(/ /, $buffer))[0] );
     }
  }

  sub sendSMTP {
    my ($debug) = shift;
    send_to_smtp($debug, 1, @_);
  }

  sub sendSMTPnoresponse {
    my ($debug) = shift;
    send_to_smtp($debug, 0, @_);
  }
